from .tokens import UserTokenManagement, SessionTokenManagement
from .values import Tvq, Property, Annotation
from collections import defaultdict


DEFAULT_SENDER_PORT_ANONYMOUS_HTTP = '55253'
DEFAULT_SENDER_PORT_USERNAME_HTTPS = '55254'      


def chunks(l, n):
    """Yield successive n-sized chunks from l.
    From https://stackoverflow.com/a/312464/11902188
    """
    for i in range(0, len(l), n):
        yield l[i:i + n]


class CanarySender(SessionTokenManagement, UserTokenManagement):
    
    __slots__ = ('_lastStoredTags')

    def __init__(self, 
                 httpPort =DEFAULT_SENDER_PORT_ANONYMOUS_HTTP, 
                 httpsPort=DEFAULT_SENDER_PORT_USERNAME_HTTPS, 
                 **configuration):
        """Canary Sender interface for pushing data to the historian service(s).

        Use with a context manager for automatic token management.
        For example, to talk to the default localhost historian using anonymous access:

            with CanarySender() as send:
                send.storeData(tvqs)

        Args:
            httpPort: (str) Port configured for anonymous HTTP access
            httpsPort: (str) Port configured for user/pass HTTPS access
        """
        super().__init__(httpPort =httpPort, httpsPort=httpsPort, **configuration)


    # Storage - File options

    def createNewFile(self, dataset, timestamp):
        """Create a new file for the selected dataset, marked at the given timestamp.
        Like fileRollover, but the tags for the file will only show those in 
          this particular file.

        Args:
            dataset: (str) Name of the dataset
            timestamp: (str) Datetime label for new file
        """
        jsonData = {
            "userToken":self.userToken,
            "sessionToken":self.sessionToken, 
            "dataset":dataset,
            "fileTime":timestamp
        }
        self._post('createNewFile', jsonData)

        
    def fileRollover(self, dataset, timestamp):
        """Create a new rolled over file labeled at the given timestamp.
        Like createNewFile, but ensures that previous file's tags are
          available, even if no data has been inserted for this file.

        Args:
            dataset: (str) Name of the dataset
            timestamp: (str) Datetime label for new file
        """
        jsonData = {
            "userToken":self.userToken,
            "sessionToken":self.sessionToken, 
            "dataset":dataset,
            "fileTime":timestamp
        }
        self._post('fileRollOver', jsonData)
    

    # Storage - Configuration

    def configureTags(self, tags):
        """Initialize parameters on the given tags.
        Each entry in the tags dict is a dict of either/both of 
            transform: (str) calculation to apply on incoming data before storing
            normalize: (str) time span to snap incoming values to.

        Args:
            tags: (dict) tags and their dict of properties and values
        """
        jsonData = {
            "userToken": self.userToken,
            "sessionToken": self.sessionToken,
            "tags": tags 
        }
        self._post("configureTags", jsonData)



    # Storage - Data
    
    def storeData(self, tvqs={}, properties={}, annotations={}, maxPageSize=25000):
        """Store data in the historian.
        Three types of data may be inserted, values, tag properties, and annotations.
        Each entry is a dictionary of tag path keys and arrays of arrays values.
        The value tuples/lists for each is:
            tvqs: (timestamp, value, quality (opt))
            properties: (prop name, timestamp, value, quality (opt)) 
            annotations: (user, timestamp, value, createdAt (opt))

        The maxPageSize limits the size of very large payloads to Canary. If the
          number of values for each tag in each dict totals more than this, then
          the update will be broken up into smaller chunks of maxPageSize each.

        Args:
            tvqs: (dict of lists) Values for tags
            properties: (dict of lists) Properties for tags
            annotations: (dict of lists) Annotations for tags
            maxPageSize: (int) Number of values to send in a single update payload        """
        
        # Make sure that we send a proper list of lists.
        def coerceList(someList):
            allowedTypes = (Tvq, Property, Annotation)
            # If the list isn't empty, check if it's a singleton entry
            if someList and not isinstance(someList[0], (tuple, list) + allowedTypes):
                return [someList.values(iso8601=True) 
                            if isinstance(someList, allowedTypes) 
                            else someList
                        ]
            else:
                return [value.values(iso8601=True) 
                            if isinstance(value, allowedTypes) 
                            else value
                        for value in values
                        ]

        # Convert from helpers objects, if needed. (This ensures it serializes correctly)
        for tag,values in tvqs.items():
            tvqs[tag] = coerceList(values)

        for tag,values in properties.items():
            properties[tag] = coerceList(values)

        for tag,values in annotations.items():
            annotations[tag] = coerceList(values)

        dataToSend = {
            'tvqs': tvqs,
            'properties': properties,
            'annotations': annotations
        }

        if sum(sum(len(values) for values in tagDict.values())
               for tagDict in dataToSend.values()
               ) < maxPageSize:
            self._storeData(**dataToSend)
        else:
            # Page out the data. Sort to make ensure earliest entries are added first
            pageLen = 0
            pageDict = {}

            for entryType in ('tvqs', 'properties', 'annotations'):
                pageDict[entryType] = tagDict = {}

                for tag,values in dataToSend[entryType].items():
                    tagDict[tag] = []

                    for value in values:
                        tagDict[tag].append(value)
                        pageLen += 1

                        if pageLen >= maxPageSize:
                            self._storeData(**pageDict)

                            # reset the page counters
                            pageDict[entryType] = tagDict = {}
                            tagDict[tag] = []
                            pageLen = 0
            else:
                if pageLen:
                    self._storeData(**pageDict)

        self._lastStoredTags = set(tvqs.keys()) | set(properties.keys()) | set(annotations.keys())


    def _storeData(self, tvqs={}, properties={}, annotations={}):
        """The call that executes storeData."""
        dataEntry = {
            "userToken":self.userToken,
            "sessionToken":self.sessionToken, 
        }
        if any((tvqs, properties, annotations)):
            if tvqs:
                dataEntry['tvqs'] = tvqs
            if properties:
                dataEntry['properties'] = properties
            if annotations:
                dataEntry['annotations'] = annotations
            self._post('storeData', dataEntry)


    def noData(self, tags=[]):
        """Adds a "No Data" entry to the end of the tags provided.
        Useful for braketting and marking a separtion chunks of tag data.

        Args:
            tags: (list) Tag paths to append to
        """
        if not tags:
            assert self._lastStoredTags, "Can't set 'No Data' without context. No tags given and _lastStoredTags is empty."
            tags = self._lastStoredTags

        dataEntry = {
            "userToken":self.userToken,
            "sessionToken":self.sessionToken,
            "tags": self._coerceToList(tags)
        }
        self._post('noData', dataEntry)


    # Info services

    def version(self):
        """Get the version of the sender service.

        Returns:
            - (str) version label 
        """
        jsonData = {
            "userToken": self.userToken
        }
        return self._singlePost("version", jsonData, "version")

    def compatibleVersion(self):
        """Get the compatible version of the sender service.
        
        Returns:
            - (str) API version sender service is compatible with
        """
        jsonData = {
            "userToken": self.userToken
        }
        return self._singlePost("compatibleVersion", jsonData, "compatibleVersion")

    def getDatasets(self, historian):
        """Get the datasets in a historian.

        Args:
            historian: (str) historian to list datasets for

        Returns:
            - (list) dataset names in historian
        """
        jsonData = {
            "userToken": self.userToken,
            "historian": historian
        }
        return self._singlePost("getDatasets", jsonData, "datasets")


"""
	A Python interface to the Canary REST API

"""

name = 'birdsong'

from .view import CanaryView
from .sender import CanarySender
from .values import Tvq, Property, Annotation


__all__ = ['CanaryView', 'CanarySender', 'Tvq', 'Property', 'Annotation']


import ciso8601, arrow
from datetime import datetime


class BaseValue(object):

    __slots__ = ('_tuple')

    _fields = ('value')
    _optional = (True)

    _timeFormat = None


    def __init__(self, value=None):
        self._tuple = (value,)
    
    @classmethod
    def keys(cls):
        return cls._fields

    def values(self, iso8601=False):
        return self._astuple(iso8601)

    def _astuple(self, iso8601=False):
        if iso8601:
            return tuple(value.isoformat() 
                            if isinstance(value, (datetime,arrow.Arrow)) 
                            else value
                         for value,optional 
                         in zip(self._tuple, self._optional) 
                         if not optional or not (value is None))
        else:
            return tuple(value 
                         for value,optional 
                         in zip(self._tuple, self._optional) 
                         if not optional or not (value is None))

    def _asdict(self,iso8601=False):
        return dict(zip(self._fields, self._astuple(iso8601)))

    def __getitem__(self, key):
        try:
            return self._tuple[key]
        except(TypeError, IndexError):
            return self._tuple[self._ixLookup[key]]

    def __iter__(self):
        return iter(self._astuple())

    def __repr__(self):
        return repr(self._asdict(iso8601=True))

    # Some time helper / coersion bits.
    # Don't try to need this: just use ISO8601 for your date format like Canary does:
    #  YYYY-MM-DD HH:mm:ss.SSSSSSZ
    def setTimeFormat(self, formatString):
        self._timeFormat = formatString

    def _coerceTimestamp(self, timestamp):
        if self._timeFormat:
            return arrow.get(timestamp, self._arrowTimeFormat)
        if isinstance(timestamp, str):
            try: # the iso8601 format first
                return arrow.Arrow.fromdatetime(ciso8601.parse_datetime(timestamp))
            except ValueError:
                raise ValueError('%r attempted to parse "%s" without a time format' % (self, timestamp))
        
        # if it's already a datetime converto to this more convenient class
        elif isinstance(timestamp, datetime):
            return arrow.Arrow.fromdatetime(timestamp)

        # if it's a tuple (like what would be used ot initialize a datetime), then init
        elif isinstance(timestamp, (tuple,list)):
            return arrow.Arrow(*timestamp)

        else:
            try: # see if arrow can convert it anyhow (like ms since epoch...)
                return arrow.get(timestamp)
            except ValueError:
                raise ValueError('%r attempted to parse "%s" without a time format' % (self, timestamp))


def _finalize(BVClass, aliases=None):
    setattr(BVClass, '_ixLookup', dict((field,ix) 
                                       for ix,field 
                                       in enumerate(BVClass._fields) ) )
    for ix,key in enumerate(BVClass._fields):
        setattr(BVClass, key, property(lambda self, ix=ix: self._tuple[ix]))
        setattr(BVClass, 'get%s' % key.capitalize(), lambda self, ix=ix: self._tuple[ix])
    if aliases:
        for ix,key in enumerate(aliases):
            setattr(BVClass, key, property(lambda self, ix=ix: self._tuple[ix]))


class Tvq(BaseValue):
    _fields = ('timestamp', 'value', 'quality')
    _optional = (False, False, True)

    def __init__(self, timestamp, value, quality=None):
        timestamp = self._coerceTimestamp(timestamp)

        self._tuple = (timestamp, value, quality)
    
_finalize(Tvq, 't v q'.split())


class Property(BaseValue):
    _fields = ('name', 'timestamp', 'value', 'quality')
    _optional = (False, False, False, True)

    def __init__(self, name, timestamp, value, quality=None):
        timestamp = self._coerceTimestamp(timestamp)

        self._tuple = (name, timestamp, value, quality)

_finalize(Property, 'n t v q'.split())


class Annotation(BaseValue):
    _fields = ('user', 'timestamp', 'value', 'createdAt')
    _optional = (False, False, False, True)

    def __init__(self, user, timestamp, value, createdAt=None):
        timestamp = self._coerceTimestamp(timestamp)
        if createdAt:
            createdAt = self._coerceTimestamp(createdAt)
        self._tuple = (user, timestamp, value, createdAt)

_finalize(Annotation, 'u t v c'.split())
    

VALUE_TYPE_MAP = {'tvq': Tvq, 'property': Property, 'annotation': Annotation}
   
def createValue(valueType='tvq', *values):
    return VALUE_TYPE_MAP[valueType.lower()](*values)
